import asyncio
import inspect

import aiocells.aio as aio

REPEATER = "cells.flow.repeater"


def repeat(function):
    if not inspect.iscoroutinefunction(function):
        raise ValueError("Event source must be a coroutine function")
    setattr(function, REPEATER, True)
    return function


def is_repeater(function):
    return getattr(function, REPEATER, False)


async def compute_flow(graph):
    assert graph is not None
    input_nodes = graph.input_nodes
    callables, running_tasks = aio.prepare_ready_set(input_nodes)
    if len(callables) > 0:
        raise Exception("Input nodes must be coroutines")
    # Wait for at least one input node to complete
    while len(running_tasks) > 0:
        completed_tasks, running_tasks = await asyncio.wait(
            running_tasks,
            return_when=asyncio.FIRST_COMPLETED
        )
        aio.raise_task_exceptions(completed_tasks)
        completed_coro_functions = [
            task.aio_coroutine_function
            for task in completed_tasks
            if is_repeater(task.aio_coroutine_function)
        ]
        callables, new_tasks = aio.prepare_ready_set(completed_coro_functions)
        assert len(callables) == 0
        running_tasks |= new_tasks
        for node in graph.topological_ordering:
            if node in input_nodes:
                continue
            if inspect.iscoroutinefunction(node):
                raise Exception("Cannot handle async internal nodes yet")
            assert callable(node)
            node()



#!/usr/bin/env python3

import asyncio
from functools import partial

import aiocells

# This example demonstrates graph nodes that are coroutines. We use
# a different computer; one that know how to deal with coroutines.


def main():
    graph = aiocells.DependencyGraph()

    # First, we add a lambda function
    before_sleep = graph.add_node(lambda: print("Sleeping..."))

    # Second, we create a coroutine function using functools.partial. This
    # is the closest we can get to a lambda for an async function
    sleep_2 = partial(asyncio.sleep, 2)

    # Finally, another lambda function
    wake_up = graph.add_node(lambda: print("Woke up!"))

    # Here, 'sleep' will implicitly be added to the graph because it is
    # part of the precedence relationship
    graph.add_precedence(before_sleep, sleep_2)
    graph.add_precedence(sleep_2, wake_up)

    # Here, we use the `async_compute_sequential`, which, like
    # `compute_sequential`, call the nodes in a topologically correct sequence.
    # However, whereas `compute_sequential` only supports vanilla callables,
    # `async_compute_sequential` additionally supports coroutine functions,
    # as defined by `inspect.iscoroutinefunction`. However, the execution is
    # still sequential. Each coroutine function is executed using 'await' and
    # must complete before the next node is executed. The function
    # `async_compute_sequential` is a coroutine and must be awaited.  Here,
    # we simply pass it to `asyncio.run`.
    asyncio.run(aiocells.async_compute_sequential(graph))



#!/usr/bin/env python3

import asyncio

import aiocells
import aiocells.demo_6 as demo_6


def main():
    stopwatch = aiocells.Stopwatch()
    graph = demo_6.create_graph(stopwatch)

    # Here, we run the same graph as the previous demo but we use
    # 'async_compute_concurrent' which will run the two sleeps concurrently.
    # Thus, the execution time will be around 2 seconds, the maximum of
    # the two sleeps.

    print("Running previous demo's graph concurrently.")
    print("Total execution time should be about 2 seconds...")
    asyncio.run(aiocells.async_compute_concurrent(graph))
    print("Computation with `async_compute_concurrent` took"
          f" {stopwatch.elapsed_time()}")



#!/usr/bin/env python3

import time

import aiocells


def main():
    graph = aiocells.DependencyGraph()
    graph.add_node(lambda: time.sleep(2))
    print("This computation will take about 2 seconds because of the sleep")
    aiocells.compute_sequential(graph)



#!/usr/bin/env python3

import asyncio
import functools
import logging

import aiocells


logger = logging.getLogger()


def main():

    clock = aiocells.ModClock()
    graph = aiocells.DependencyGraph()

    # Three completely unrelated sequences are added to the graph. They
    # run concurrently.
    time_0 = aiocells.ModVariable(clock)
    timer_0 = functools.partial(aiocells.timer, 0, time_0)
    printer_0 = aiocells.ModPrinter(clock, time_0, "time_0 changed to {value}")
    graph.add_precedence(timer_0, time_0)
    graph.add_precedence(time_0, printer_0)

    time_1 = aiocells.ModVariable(clock)
    timer_1 = functools.partial(aiocells.timer, 1, time_1)
    printer_1 = aiocells.ModPrinter(clock, time_1, "time_1 changed to {value}")
    graph.add_precedence(timer_1, time_1)
    graph.add_precedence(time_1, printer_1)

    time_3 = aiocells.ModVariable(clock)
    timer_3 = functools.partial(aiocells.timer, 3, time_3)
    printer_3 = aiocells.ModPrinter(clock, time_3, "time_3 changed to {value}")
    graph.add_precedence(timer_3, time_3)
    graph.add_precedence(time_3, printer_3)

    # With a flow computation, only input nodes (those with no dependencies),
    # can be coroutine functions. When any one of those nodes returns, the
    # graph is computed. When this happens, we are generally only interested
    # in nodes that change as a result of the input node returning. So, in
    # this case, we see a message at 0 seconds, 1 second and 3 seconds.
    asyncio.run(aiocells.compute_flow(graph))



#!/usr/bin/env python3

import asyncio
import functools
import logging

import aiocells


logger = logging.getLogger()

# Demonstrates that the same graph can be computed twice


def main():

    graph = aiocells.DependencyGraph()

    time = aiocells.Variable()

    # 'aio.timer' will put the current time in the 'time' variable when
    # one second has expired
    timer = functools.partial(aiocells.timer, 1, time)
    printer = aiocells.Printer(time, "variable changed to {value}")

    graph.add_precedence(timer, time)
    graph.add_precedence(time, printer)
    logger.debug("graph: %s", graph)

    logger.info("First computation...")
    asyncio.run(aiocells.async_compute_concurrent(graph))
    logger.debug("graph: %s", graph)

    logger.info("Second computation...")
    asyncio.run(aiocells.async_compute_concurrent(graph))
    logger.debug("graph: %s", graph)
